"""\
========
Analysis
========
The :class:`~ENPMDA.analysis.base.DaskChunkMdanalysis` class is
the base class to define multi-frame parallel analysis for
the MD trajectories. It functions as a wrapper of the
MDAnalysis analysis functions to map the analysis to the Dask dataframe.
This class takes care of loading the right universe and dumping the
results as a `npy` file to avoid huge memory footprint
and dask scheduler clogging. 

To define a new analysis,
:class:`~ENPMDA.analysis.base.DaskChunkMdanalysis`
needs to be subclassed.
`run_analysis` need to be defined. ``name`` will be the feature name
appending to the dataframe. In default, only protein universe file
will be used to run analysis. It can be overridden by defining
``universe_file=system``. Some information of the feature can also be
provided and stored in ``_feature_info``::

    from ENPMDA.analysis import DaskChunkMdanalysis
    class NewAnalysis(DaskChunkMdanalysis):
        name = 'new_analysis'
        universe_file = 'protein'

        def get_feature_info(self, universe):
            return ['some_info']

        def run_analysis(self, universe, start, stop, step):
            result = []
            for ts in universe.trajectory[start:stop:step]:
                result.append(some_analysis(universe.atoms))
            return result

Classes
=======
.. autoclass:: DaskChunkMdanalysis
   :members:
"""
import os
import dask
import numpy as np
import uuid
import pickle
import itertools


class AnalysisResult(dict):
    """
    This class is used internally to store the results of the analysis.
    """

    def __init__(self,
                 dd_dataframe,
                 dataframe,
                 working_dir,
                 timestamp):
        super().__init__()

        self.dd_dataframe = dd_dataframe
        self.dataframe = dataframe
        self.working_dir = working_dir

        self.timestamp = timestamp

        os.makedirs(self.filename, exist_ok=True)

    def compute(self, item=None):
        if item is None:
            for item, df in self.items():
                if isinstance(df, dask.dataframe.core.DataFrame):
                    self[item] = df.compute()[['system', item]]
        elif item in self.keys():
            if isinstance(df, dask.dataframe.core.DataFrame):
                self[item] = df.compute()[['system', item]]
        else:
            raise ValueError(item + ' not in dict')

    def add_column_to_results(self, analysis_function, **kwargs):
        kwargs['filename'] = self.filename

        # sanity check and get feature info
        check_analysis_function = analysis_function(**kwargs)
        if check_analysis_function.universe_file == 'protein':
            universe = pickle.load(open(self.dataframe.iloc[0].universe_protein, "rb"))
        else:
            universe = pickle.load(open(self.dataframe.iloc[0].universe_system, "rb"))
        
        feature_info = check_analysis_function.set_feature_info(universe)
        check_result = np.asarray(check_analysis_function.run_analysis(universe, 0, 2, 1))
        check_result = check_result.reshape(check_result.shape[0], -1)
        if check_result.shape[1] != len(feature_info):
            raise ValueError(f'The shape of feature info ({len(feature_info)}) '
            f'does not match the shape of analysis ({check_result.shape[1]}).')
        if check_result.shape[0] != 2:
            raise ValueError('The len of the result'
            'does not match the number of the frame.'
            'Hint: run_analysis should return a list'
            'with the shape as the trajectory frames.')
        np.save(check_analysis_function.feature_info_loc, feature_info)

        item_name = analysis_function.name
        meta = dict(self.dd_dataframe.dtypes)
        meta[item_name] = "f8"

        # add analysis to dataframe
        self[item_name] = self.dd_dataframe.map_partitions(lambda df:
                                                      df.assign(
                                                          **{item_name: analysis_function(**kwargs)(df)}),
                                                      meta=meta,
                                                      token=item_name
                                                      ).persist()


    def append_to_dataframe(self, dataframe):
        for item in self.keys():
            dataframe[item] = self[item].iloc[:, 1]

    @property
    def filename(self):
        return self.working_dir + '/analysis_results/' + self.timestamp + '/'


class DaskChunkMdanalysis(object):
    """
    This class is the base class for all analysis classes.
    The analysis results will be dumped as a `npy` file with
    unique uuid for each partition.
    """

    name = 'analysis'
    universe_file = 'protein'

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

        # get unique uuid for each partition
        self._partition = uuid.uuid4().hex
        self.feature_info = []

    def __call__(self, df):
        return self.run_df(df)

    def run_df(self, df):
        result = []

        # run analysis
        for system, df_sys in df.groupby(['system']):
            # if system information has to be used set `universe_file =
            # 'system'`
            if self.universe_file == 'protein':
                universe = pickle.load(open(df_sys.universe_protein.iloc[0], "rb"))
            else:
                universe = pickle.load(
                    open(df_sys.universe_system.iloc[0], "rb"))
            start = df_sys.frame.iloc[0]
            stop = df_sys.frame.iloc[-1] + 1
            step = df_sys.stride.iloc[0]
            start, stop, step = universe.trajectory.check_slice_indices(
                start, stop, step)

            result.append(self.run_analysis(universe, start, stop, step))
            del universe

        result_ordered = list(itertools.chain.from_iterable(result))
        # store results and only return the location of the files
        np.save(self.location, result_ordered)
        n_result = len(result_ordered)
        del result_ordered
        return n_result * [self.location]

#        return list(itertools.chain.from_iterable(result))

    def set_feature_info(self, universe):
        """
        This function is used to set the feature information.
        Shold return a list of features.
        """
        raise NotImplementedError('Only for inheritance.')

    def run_analysis(self, universe, start, stop, step):
        """
        The function to be overwritten by the analysis class.
        """
        raise NotImplementedError('Only for inheritance.')

    @property
    def name(self):
        return self._name

    @property
    def partition(self):
        return str(self._partition)

    @property
    def location(self):
        return self.filename + self.name + '_' + self.partition + '.npy'

    @property
    def feature_info_loc(self):
        return self.filename + self.name + '_feature_info.npy'


from pkg_resources import resource_filename

ensemble_ala_tpr = []
ensemble_ala_traj = []
ensemble_ala_top = []

for rep in range(1, 9):
    ensemble_ala_tpr.append(resource_filename(__name__,
                                              f'datafile/ensemble_AlaDipeptide/rep{rep}/md.tpr'))
    ensemble_ala_traj.append(resource_filename(__name__,
                                               f'datafile/ensemble_AlaDipeptide/rep{rep}/md.xtc'))
    ensemble_ala_top.append(resource_filename(__name__,
                                              f'datafile/ensemble_AlaDipeptide/rep{rep}/start.pdb'))



# From https://groups.google.com/g/mdnalysis-discussion/c/umDpvbCmQiE/m/FKtNClazAwAJ
# Author: Richard Gowers

from MDAnalysis.transformations.base import TransformationBase
import numpy as np


class GroupHug(TransformationBase):
    def __init__(self, center, *others):
        super().__init__(max_threads=1,
                         parallelizable=True)
        self.c = center
        self.o = others

    @staticmethod
    def calc_restoring_vec(ag1, ag2):
        box = ag1.dimensions[:3]
        dist = ag1.center_of_mass() - ag2.center_of_mass()

        return box * np.rint(dist / box)

    def _transform(self, ts):
        # loop over other atomgroups and shunt them into nearest image to
        # center
        for i in self.o:
            rvec = self.calc_restoring_vec(self.c, i)
            i.translate(+rvec)
        return ts



"""\
===========
MDDataFrame
===========
The :class:`~ENPMDA.MDDataFrame` class both store
the metadata of simulations in the ensemble and functions as
a dask dataframe to add, compute, and store analysis.

A ``MDDataFrame`` is created from files::

    from ENPMDA import MDDataFrame
    md_dataframe = MDDataFrame()
    md_dataframe.add_traj_ensemble(traj_ensemble, npartitions=16)


Classes
=======
.. autoclass:: MDDataFrame
   :members:
"""


from datetime import datetime
import warnings
import numpy as np
import dask.dataframe as dd
import dask
import pandas as pd
import MDAnalysis as mda
import os
import pickle
import shutil

from ENPMDA.analysis.base import AnalysisResult
from ENPMDA.preprocessing import TrajectoryEnsemble

timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
meta_data_list = ["universe_protein",
                  "universe_system",
                  "system",
                  "traj_name",
                  "frame",
                  "traj_time",
                  "stride",
                  ]


class MDDataFrame(object):
    r"""
    Class to store the metadata and analysis results
    of the ensemble simulations.

    It uses pandas.DataFrame to store metadata
    and dask.DataFrame to distribute computation jobs
    so that the parallel analysis can be performed not
    only for one trajectory but also across simulations
    and analyses.
    """

    def __init__(self,
                 dataframe_name,
                 meta_data_list=meta_data_list,
                 timestamp=timestamp):
        """
        Parameters
        ----------
        dataframe_name: str
            The name of the dataframe
            It will be used as the folder to save
            all the analysis results. 
            It can also be the absolute path to the folder.
        meta_data_list: list, optional
            List of metadata in the dataframe.
            In default, the locations of pickled universes
            of protein and system, the system index, the
            trajectory filename, the frame index, the
            trajectory time, and the stride are stored.
        timestamp: str, optional
            The timestamp of creating the ensemble
            It will be set to the current time if not provided.
        """
        self.dataframe_name = dataframe_name
        self.dataframe = pd.DataFrame(
            columns=meta_data_list
        )
        self.computed = False
        if not os.path.isabs(self.dataframe_name):
            self.working_dir = os.getcwd() + '/'
        else:
            self.working_dir = ''
        self.timestamp = timestamp
        self.trajectory_ensemble = None
        self.analysis_list = []

    def add_traj_ensemble(self,
                          trajectory_ensemble: TrajectoryEnsemble,
                          npartitions,
                          stride=1
                          ):
        """
        Parameters
        ----------
        trajectory_ensemble: ENPMDA.TrajectoryEnsemble
            The trajectory ensemble to be added to the
            dataframe.
        npartitions: int
            The number of partitions to be used in
            the dask dataframe.
        stride: int, optional
            The stride to be used in the dask dataframe.
            It is used to skip frames in the trajectory.
        """
        if self.trajectory_ensemble is not None:
            raise ValueError('Trajectory ensemble already added')

        self.trajectory_ensemble = trajectory_ensemble

        if trajectory_ensemble.protein_trajectory_files is None:
            warnings.warn('The provided trajectory ensemble '
            'only contain raw trajectories '
            'all analysis will be performed on the raw trajectories',
            stacklevel=2)
            self.trajectory_files = trajectory_ensemble.trajectory_files
            self.protein_trajectory_files = trajectory_ensemble.trajectory_files
            self.system_trajectory_files = trajectory_ensemble.trajectory_files
        else:
            self.trajectory_files = trajectory_ensemble.trajectory_files
            self.protein_trajectory_files = trajectory_ensemble.protein_trajectory_files
            self.system_trajectory_files = trajectory_ensemble.system_trajectory_files

        self.npartitions = npartitions
        self.stride = stride

        meta_data_jobs = []
        for ind, trajectory in enumerate(self.protein_trajectory_files):
            meta_data_jobs.append(
                dask.delayed(self._append_metadata)(
                    trajectory, system=ind))

        meta_data = dask.compute(meta_data_jobs)[0]

        for i, trajectory in enumerate(self.protein_trajectory_files):
            self.dataframe = pd.concat([self.dataframe,
                                        pd.DataFrame(meta_data[i],
                                                     columns=self.dataframe.columns)],
                                       ignore_index=True
                                       )

        self.dataframe.frame = self.dataframe.frame.apply(int)
        self.dataframe.traj_time = self.dataframe.traj_time.apply(float)

        self._init_analysis_results(npartitions=self.npartitions)

    def _append_metadata(self, universe, system):
        universe_system = self.system_trajectory_files[system]

        u = pickle.load(open(universe, "rb"))
        u_sys = pickle.load(open(universe_system, "rb"))
        if u.trajectory.n_frames != u_sys.trajectory.n_frames:
            raise ValueError(
                f'In system {system}, number of frames in protein and system trajectories are different!')
        rep_data = []

        md_name = u.trajectory.filename
        timestep = u.trajectory.dt

        for i in range(0, u.trajectory.n_frames, self.stride):
            rep_data.append([universe,
                             universe_system,
                             system,
                             md_name,
                             i,
                             i * timestep,
                             self.stride
                             ])
        del u
        return rep_data

    def _init_analysis_results(self, npartitions):
        self.dd_dataframe = dd.from_pandas(self.dataframe,
                                           npartitions=npartitions)
        print('Dask dataframe generated with {} partitions'.format(npartitions))
        self.analysis_results = AnalysisResult(self.dd_dataframe,
                                               self.dataframe,
                                               working_dir=self.filename,
                                               timestamp=self.timestamp)

    def add_analysis(self, analysis, overwrite=False):
        """
        Add an analysis to the dataframe.

        Parameters
        ----------
        analysis: ENPMDA.analysis.base.DaskChunkMdanalysis
            The analysis to be added to the dataframe.
        overwrite: bool, optional
            Whether to overwrite the analysis if it is
            already in the dataframe.
        """
        if analysis.name in self.analysis_list and not overwrite:
            warnings.warn(f'Analysis {analysis.name} already added, add overwrite=True to overwrite',
                          stacklevel=2)
        elif analysis.name in self.analysis_list and overwrite:
            warnings.warn(f'Analysis {analysis.name} already added, overwriting!',
                          stacklevel=2)
            self.analysis_list.remove(analysis.name)
            self.analysis_list.append(analysis.name)
            self.analysis_results.add_column_to_results(analysis)
        else:
            self.analysis_list.append(analysis.name)
            self.analysis_results.add_column_to_results(analysis)

    def compute(self):
        """
        Compute the analysis results.
        It will be append the analysis results to
        the dataframe.
        """
        if not self.computed:
            self.analysis_results.compute()
        self.analysis_results.append_to_dataframe(self.dataframe)
        self.computed = True


    def get_feature(self, feature_list):
        """
        Get the features from the dataframe.

        Parameters
        ----------
        feature_list: list of str
            The list of features to be extracted.
        """
        if not self.computed:
            self.compute()

        feature_dataframe = self.dataframe[['system', 'traj_name', 'frame', 'traj_time']].copy()
        for feature in feature_list:
            raw_data = np.concatenate([np.load(location,
                                        allow_pickle=True)
                    for location, df in self.dataframe.groupby(feature,
                    sort=False)])
            raw_data =  raw_data.reshape(raw_data.shape[0], -1)
            feat_info = np.load(self.analysis_results.filename + feature + '_feature_info.npy')
            feature_dataframe[feat_info] = raw_data
        return feature_dataframe


    def save(self, filename='dataframe'):
        """
        Compute the analysis results and
        save the dataframe to a pickle file.
        
        Parameters
        ----------
        filename: str, optional
            The name of the pickle file.
            It will be saved in the working directory
        """

        if not self.computed:
            self.compute()

        if not os.path.exists(f'{self.filename}{filename}.pickle'):
            with open(f'{self.filename}{filename}.pickle', 'wb') as f:
                pickle.dump(self.dataframe, f)
        else:
            md_data_old = pickle.load(
                open(f'{self.filename}{filename}.pickle', 'rb'))

            if set(md_data_old.universe) != set(self.dataframe.universe):
                print('New seeds added')
                with open(f'{self.filename}{filename}.pickle', 'wb') as f:
                    pickle.dump(self.dataframe, f)
            elif md_data_old.shape[0] != self.dataframe.shape[0]:
                print('N.frame changed')

                old_cols = md_data_old.columns
                new_cols = self.md_data.columns
                print('New: ' + np.setdiff1d(new_cols, old_cols))

                extra_cols = np.setdiff1d(new_cols, old_cols)

                for extra_col in extra_cols:
                    md_data_old[extra_col] = self.md_data[extra_col]

                print('Common: ' + np.intersect1d(new_cols, old_cols))
                common_cols = np.intersect1d(new_cols, old_cols)

                for common_col in common_cols:
                    md_data_old[common_col] = self.md_data[common_col]

                shutil.copyfile(
                    f'{self.filename}{filename}.pickle',
                    f'{self.filename}{filename}_{self.timestamp}.pickle')
                md_data_old.to_pickle(f'{self.filename}{filename}.pickle')
            else:
                print('No changes')
                with open(f'{self.filename}{filename}.pickle', 'wb') as f:
                    pickle.dump(self.dataframe, f)

        with open(f'{self.filename}{filename}_md_dataframe.pickle', 'wb') as f:
            pickle.dump(self, f)

    @property
    def filename(self):
        """
        The saving location of all the pickled files.
        """
        return os.path.abspath(self.working_dir + self.dataframe_name) + '/'



import numpy as np
import itertools

from MDAnalysis.analysis.rms import RMSD
from MDAnalysis.analysis.distances import distance_array
from MDAnalysis.lib.distances import calc_bonds
from MDAnalysis.analysis.distances import self_distance_array
from MDAnalysis.analysis.dihedrals import Dihedral as mda_dihedral

from .base import DaskChunkMdanalysis


class get_backbonetorsion(DaskChunkMdanalysis):
    name = 'torsion'

    def set_feature_info(self, universe):
        feature_info = []
        res_phi_list = [x.resindices[0]
                        for x in universe.atoms.residues.phi_selections() if x is not None]

        res_psi_list = [x.resindices[0]
                        for x in universe.atoms.residues.psi_selections() if x is not None]

        feature_info.extend(['_'.join([str(residex), ang]) for residex, ang in
                                   itertools.product(res_phi_list,
                                                     ['phi_cos', 'phi_sin'])])
        feature_info.extend(['_'.join([str(residex), ang]) for residex, ang in
                                   itertools.product(res_psi_list,
                                                     ['psi_cos', 'psi_sin'])])

        return feature_info

    def run_analysis(self, universe, start, stop, step):
        phi_ag = [x for x in universe.atoms.residues.phi_selections()
                  if x is not None]
        phi_angle_ana = mda_dihedral(phi_ag).run(start, stop, step)
        phi_angle = np.concatenate([[np.cos(np.deg2rad(phi_angle_ana.results['angles']))],
                                    [np.sin(np.deg2rad(phi_angle_ana.results['angles']))]]).transpose(1, 0, 2)

        psi_ag = [x for x in universe.atoms.residues.psi_selections()
                  if x is not None]
        psi_angle_ana = mda_dihedral(psi_ag).run(start, stop, step)
        psi_angle = np.concatenate([[np.cos(np.deg2rad(psi_angle_ana.results['angles']))],
                                    [np.sin(np.deg2rad(psi_angle_ana.results['angles']))]]).transpose(1, 0, 2)

        torsion_angle = np.concatenate([psi_angle, phi_angle], axis=1)
        torsion_angle = torsion_angle.reshape(
            torsion_angle.shape[0],
            torsion_angle.shape[1] *
            torsion_angle.shape[2])

        return torsion_angle


class get_atomic_position(DaskChunkMdanalysis):
    name = 'at_position'

    def set_feature_info(self, universe):
        backbone_atoms = universe.select_atoms('backbone')
        return ['_'.join([str(residex), feat]) for residex, feat in itertools.product(
            backbone_atoms.resindices, ['_bb_pos_x', '_bb_pos_y', '_bb_pos_z'])]

    def run_analysis(self, universe, start, stop, step):
        backbone_atoms = universe.select_atoms('backbone')
        result = []
        for ts in universe.trajectory[start:stop:step]:
            result.append(
                (backbone_atoms.positions -
                 backbone_atoms.center_of_geometry()).ravel())

        return result


class get_rmsd_init(DaskChunkMdanalysis):
    name = 'rmsd_to_init'

    def set_feature_info(self, universe):
        return ['rmsd_to_init']

    def run_analysis(self, universe, start, stop, step):

        name_backbone = universe.select_atoms('backbone')
        rmsd = RMSD(name_backbone, name_backbone).run(start, stop, step)
        return rmsd.results['rmsd'].T[2]


class get_protein_hydration(DaskChunkMdanalysis):
    name = 'protein_hydration'
    universe_file = 'system'

    def set_feature_info(self, universe):
        return ['protein_hydration']

    def run_analysis(self, universe, start, stop, step):

        prot_hydration = universe.select_atoms(
            "around 5 backbone", updating=True)

        n_hydration = []
        for ts in universe.trajectory[start:stop:step]:
            n_hydration.append(prot_hydration.n_atoms)
        return n_hydration


#!/usr/bin/env python

"""Tests for `ENPMDA` package."""
import os
import tempfile
import pytest
import numpy as np
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_equal,
    assert_array_equal,
)

from ENPMDA import MDDataFrame
from ENPMDA.preprocessing import TrajectoryEnsemble
from ENPMDATests.datafiles import (
    ensemble_ala_tpr,
    ensemble_ala_traj,
    ensemble_ala_top
)


class TestDDataFrameCreation(object):
    @pytest.fixture
    def tempdir(self):
        tempdir = tempfile.mkdtemp()
        return tempdir

    def test_inititialize_trajectoryensemble(self, tempdir):
        traj_ensembles = TrajectoryEnsemble(ensemble_name= './' + tempdir + '/test_traj_ensemble',
                                            topology_list=ensemble_ala_top,
                                            trajectory_list=ensemble_ala_traj,
                                            tpr_list=ensemble_ala_tpr,
                                            updating=False,
                                            only_raw=False)
        assert_equal(traj_ensembles.filename,
                     os.getcwd() + tempdir + '/' + 'test_traj_ensemble' + '/',
                     "ensemble name is not set correctly")

    def test_initialize_mddataframe(self, tempdir):
        md_dataframe = MDDataFrame(dataframe_name='./' + tempdir + '/test_init_dataframe')

        assert_equal(md_dataframe.filename,
                     os.getcwd() + tempdir + '/' + 'test_init_dataframe' + '/',
                     "working dir is not set correctly")

    def test_init_trajectoryensemble_abs_path(self, tempdir):
        traj_ensembles = TrajectoryEnsemble(ensemble_name=tempdir + '/test_traj_ensemble',
                                            topology_list=ensemble_ala_top,
                                            trajectory_list=ensemble_ala_traj,
                                            tpr_list=ensemble_ala_tpr,
                                            updating=False,
                                            only_raw=False)
        assert_equal(traj_ensembles.filename,
                     tempdir + '/' + 'test_traj_ensemble' + '/',
                     "ensemble name is not set correctly")

    def test_init_mddataframe_abs_path(self, tempdir):
        md_dataframe = MDDataFrame(dataframe_name=tempdir + '/test_init_dataframe')

        assert_equal(md_dataframe.filename,
                     tempdir + '/' + 'test_init_dataframe' + '/',
                     "working dir is not set correctly")


class TestAddTrajEnsemble(object):

    @pytest.fixture
    def tempdir(self):
        tempdir = tempfile.mkdtemp()
        return tempdir

    @pytest.fixture()
    def md_dataframe(self, tempdir):
        return MDDataFrame(dataframe_name=tempdir + '/test_init_dataframe')

    @pytest.fixture()
    def traj_ensemble(self, tempdir):
        return TrajectoryEnsemble(ensemble_name=tempdir + '/test_ensemble',
                                  topology_list=ensemble_ala_top,
                                  trajectory_list=ensemble_ala_traj,
                                  tpr_list=ensemble_ala_tpr,
                                  updating=False,
                                  only_raw=False)

    @pytest.fixture()
    def raw_traj_ensemble(self, tempdir):
        return TrajectoryEnsemble(ensemble_name=tempdir + '/test_ensemble',
                                  topology_list=ensemble_ala_top,
                                  trajectory_list=ensemble_ala_traj,
                                  tpr_list=ensemble_ala_tpr,
                                  updating=False,
                                  only_raw=True)

    def test_add_trajectory_ensemble(self, md_dataframe, traj_ensemble):
        md_dataframe.add_traj_ensemble(traj_ensemble, npartitions=10)

        assert md_dataframe.dd_dataframe is not None
        assert_array_equal(md_dataframe.dataframe.shape,
                           (168, 7),
                           "Dataframe shape is not correct")
        assert_equal(md_dataframe.dataframe.columns.tolist(),
                     ['universe_protein', 'universe_system', 'system',
                         'traj_name', 'frame', 'traj_time', 'stride'],
                     "Dataframe columns are not correct")

        assert_equal(md_dataframe.npartitions, 10,
                     "npartitions not set correctly")
        assert md_dataframe.analysis_results is not None

    def test_add_raw_trajectory_ensemble(self, md_dataframe, raw_traj_ensemble):
        md_dataframe.add_traj_ensemble(raw_traj_ensemble, npartitions=10)

        assert md_dataframe.dd_dataframe is not None
        assert_array_equal(md_dataframe.dataframe.shape,
                           (168, 7),
                           "Dataframe shape is not correct")
        assert_equal(md_dataframe.dataframe.columns.tolist(),
                     ['universe_protein', 'universe_system', 'system',
                         'traj_name', 'frame', 'traj_time', 'stride'],
                     "Dataframe columns are not correct")

        assert_equal(md_dataframe.npartitions, 10,
                     "npartitions not set correctly")
        assert md_dataframe.analysis_results is not None


import re


def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    '''
    return [atoi(c) for c in re.split(r'(\d+)', text)]



"""\
==================
TrajectoryEnsemble
==================
The :class:`~ENPMDA.preprocessing.TrajectoryEnsemble` class both store
the information of simulations in the ensemble and preload it by
serialization. It can also apply on-the-fly transformations to the
trajectories and extract selected components (e.g. only protein)
to seperated files.

A ``TrajectoryEnsemble`` is created from files::

    from ENPMDA.preprocessing import TrajectoryEnsemble
    traj_ensembles = TrajectoryEnsemble(ensemble_name='ensemble',
                                      topology_list=ensemble_top,
                                      trajectory_list=ensemble_traj)

In order to add transformations e.g. wrap/unwrap, extra ``tpr_list``
is required as input to provide bonded information.


Classes
=======
.. autoclass:: TrajectoryEnsemble
   :members:
"""

import os.path
import warnings
from datetime import datetime
import gc
import pickle
import os
import MDAnalysis as mda
import MDAnalysis.transformations as trans
import dask

from ENPMDA.utils import GroupHug

timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")


class TrajectoryEnsemble(object):
    r"""Class to store an ensemble of simulations.
    It can also be used to apply transformations to
    the trajectory and either return the raw trajectory or
    the processed trajectory for protein and system.

    Note
    ----
    Each trajectory should be stored in its own folder.

    Warning
    -------
    It will generate `protein.pdb`, `system.pdb`,
    `protein.xtc`, and `system.xtc` files
    in the same folder as the loading trajectory.
    """
    def __init__(self,
                 ensemble_name,
                 topology_list,
                 trajectory_list,
                 tpr_list=None,
                 timestamp=timestamp,
                 updating=True,
                 only_raw=False):
        r"""
        Parameters
        ----------
        ensemble_name: str
            The name of the ensemble. It will be used as the folder
            to save the pickled Universes. 
            It can also be the absolute path to the folder.

        topology_list: list
            List of topology files, e.g. gro, pdb, etc.

        trajectory_list: list
            List of trajectory files, e.g. xtc, trr, etc.

        tpr_list: list, optional
            List of tpr files. For providing extra bonded information.

        timestamp: str, optional
            The timestamp of creating the ensemble
            It will be set to the current time if not provided.

        updating: bool, optional
            If True, the trajectory will be updated
            even if the trajectory was processed before.

        only_raw: bool, optional
            If True, only the raw trajectory will be returned.
            Otherwise, both processed trajectories for protein and
            system will be returned.
        """

        if len(topology_list) != len(trajectory_list):
            raise ValueError(
                'topology_list and trajectory_list must have the same length.')
        self.ensemble_name = ensemble_name
        self.topology_list = topology_list
        self.trajectory_list = trajectory_list
        self.tpr_list = tpr_list
        self.timestamp = timestamp
        self.updating = updating
        self.only_raw = only_raw

        if self.tpr_list is None:
            self.fix_chain = False
            print('No tpr_list provided. \n',
                  'PBC and chain cannot be fixed.')
        else:
            self.fix_chain = True

        if not os.path.isabs(self.ensemble_name):
            self.working_dir = os.getcwd() + '/'
        else:
            self.working_dir = ''

        os.makedirs(self.filename, exist_ok=True)

        if self.updating or not os.path.isfile(
                self.filename + "raw_traj.pickle"):
            self._processing_ensemble()
        else:
            self.trajectory_files = pickle.load(
                open(self.filename + "raw_traj.pickle", "rb"))
        if not self.only_raw:
            if self.updating or not os.path.isfile(
                    self.filename + "protein.pickle"):
                self._processing_protein()
            else:
                self.protein_trajectory_files = pickle.load(
                    open(self.filename + "protein.pickle", "rb"))

            if updating or not os.path.isfile(self.filename + "system.pickle"):
                self._processing_system()
            else:
                self.system_trajectory_files = pickle.load(
                    open(self.filename + "system.pickle", "rb"))
        else:
            self.system_trajectory_files = None
            self.protein_trajectory_files = None
    def _processing_ensemble(self):
        load_job_list = []
        if not os.path.isfile(self.filename + "raw_traj.pickle"):
            for ind, (topology, trajectory) in enumerate(
                    zip(self.topology_list, self.trajectory_list)):
                load_job_list.append(
                    dask.delayed(
                        self._preprocessing_raw_trajectory)(
                        topology, trajectory, ind))
        else:
            for ind, (topology, trajectory) in enumerate(
                    zip(self.topology_list, self.trajectory_list)):
                if os.path.getmtime(trajectory) > os.path.getmtime(
                        self.filename + "raw_traj.pickle"):
                    print(trajectory + ' modified.')
                    load_job_list.append(
                        dask.delayed(
                            self._preprocessing_raw_trajectory)(
                            topology, trajectory, ind))
                else:
                    print(trajectory + ' on hold.')
                    load_job_list.append(
                        dask.delayed(
                            self._load_preprocessing_trajectory)(trajectory))

        self.trajectory_files = dask.compute(load_job_list)[0]
        print('dask finished')
        with open(self.filename + "raw_traj.pickle", "wb") as f:
            pickle.dump(self.trajectory_files, f)
            print('pickle raw_traj universe done')

    def _processing_protein(self):
        load_job_list = []
        for trajectory in self.trajectory_list:
            traj_path = os.path.dirname(trajectory)

            if os.path.isfile(traj_path + "/protein.xtc"):
                load_job_list.append(
                    dask.delayed(
                        self._load_protein)(trajectory))
        self.protein_trajectory_files = dask.compute(load_job_list)[0]
        print('dask finished')
        with open(self.filename + "protein.pickle", "wb") as f:
            pickle.dump(self.protein_trajectory_files, f)
            print('pickle traj protein universe done')

    def _processing_system(self):
        load_job_list = []
        for trajectory in self.trajectory_list:
            traj_path = os.path.dirname(trajectory)
            if os.path.isfile(traj_path + "/system.xtc"):
                load_job_list.append(dask.delayed(
                    self._load_system)(trajectory))
        self.system_trajectory_files = dask.compute(load_job_list)[0]
        print('dask finished')
        with open(self.filename + "system.pickle", "wb") as f:
            pickle.dump(self.system_trajectory_files, f)
            print('pickle traj system universe done')

    def _preprocessing_raw_trajectory(self, topology, trajectory, ind):
        #    print(trajectory)
        traj_path = os.path.dirname(trajectory)
        # to ignore most unnecessary warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            u = mda.Universe(topology,
                             trajectory)

            u_prot = u.select_atoms('protein')
            if self.fix_chain:
                u_bond = mda.Universe(self.tpr_list[ind])
                u.add_bonds(u_bond.bonds.to_indices())

                prot_chain_list = []
                for chain in u_prot.segments:
                    prot_chain_list.append(chain.atoms)
                non_prot = u.select_atoms('not protein')
                prot_group = GroupHug(*prot_chain_list)
                wrap = trans.wrap(non_prot)
                unwrap = trans.unwrap(u.atoms)
                center_in_box = trans.center_in_box(u_prot)

            rot_fit_trans = trans.fit_rot_trans(
                u.select_atoms('name CA'), u.select_atoms('name CA'))
            if self.fix_chain:
                u.trajectory.add_transformations(
                    *[unwrap, prot_group, center_in_box, wrap, rot_fit_trans])
            else:
                u.trajectory.add_transformations(*[rot_fit_trans])

            if not self.only_raw:
                with mda.Writer(traj_path + '/protein.xtc',
                                u.select_atoms('protein').n_atoms) as W_prot, \
                    mda.Writer(traj_path + '/system.xtc',
                               u.atoms.n_atoms) as W_sys:
                    for time, ts in enumerate(u.trajectory):
                        W_prot.write(u.select_atoms('protein'))
                        W_sys.write(u.atoms)

                u.select_atoms('protein').write(traj_path + '/protein.pdb')
                u.atoms.write(traj_path + '/system.pdb')

        with open(self.filename + '_'.join(trajectory.split('/')) + '.pickle', 'wb') as f:
            pickle.dump(u, f)

        del u
        if self.fix_chain:
            del u_bond
        gc.collect()
        return self.filename + '_'.join(trajectory.split('/')) + '.pickle'

    def _load_preprocessing_trajectory(self, trajectory):
        return self.filename + '_'.join(trajectory.split('/')) + '.pickle'

    def _load_protein(self, trajectory):
        traj_path = os.path.dirname(trajectory)
        u = mda.Universe(traj_path + '/protein.pdb',
                         traj_path + '/protein.xtc')

        with open(self.filename + '_'.join(trajectory.split('/')) + '_prot.pickle', 'wb') as f:
            pickle.dump(u, f)
        return self.filename + '_'.join(trajectory.split('/')) + '_prot.pickle'

    def _load_system(self, trajectory):
        traj_path = os.path.dirname(trajectory)
        u = mda.Universe(traj_path + '/system.pdb',
                         traj_path + '/system.xtc')

        with open(self.filename + '_'.join(trajectory.split('/')) + '_sys.pickle', 'wb') as f:
            pickle.dump(u, f)
        return self.filename + '_'.join(trajectory.split('/')) + '_sys.pickle'

    @property
    def filename(self):
        """
        The saving location of all the pickled files.
        """
        return os.path.abspath(self.working_dir + self.ensemble_name) + '/'